from click.testing import CliRunner
from spare.cli import create_cli, restore_cli, validate_cli


def test_create_cli(endpoint, access_key, secret_key, temporary_path):
    runner = CliRunner()

    result = runner.invoke(create_cli, [
        '--endpoint', endpoint.geturl(),
        '--access-key', access_key,
        '--secret-key', secret_key,
        '--password', 'foobar',
        '--bucket', 'foobar',
        '--path', str(temporary_path)
    ])
    assert result.exit_code == 0

    result = runner.invoke(restore_cli, [
        '--endpoint', endpoint.geturl(),
        '--access-key', access_key,
        '--secret-key', secret_key,
        '--password', 'foobar',
        '--bucket', 'foobar',
        '--path', str(temporary_path)
    ])
    assert result.exit_code == 0

    result = runner.invoke(validate_cli, [
        '--endpoint', endpoint.geturl(),
        '--access-key', access_key,
        '--secret-key', secret_key,
        '--password', 'foobar',
        '--bucket', 'foobar',
    ])
    assert result.exit_code == 0



import pytest
import random

from hypothesis import given, strategies
from spare.block import AVAILABLE_BLOCKS
from secrets import token_bytes


@pytest.mark.parametrize("cls", AVAILABLE_BLOCKS.values())
def test_block(cls):
    block = cls(password=b'password', nonce=token_bytes(16), data=b'foobar')

    block.encrypt()
    assert block.data != b'foobar'

    block.decrypt()
    assert block.data == b'foobar'


@pytest.mark.parametrize("cls", AVAILABLE_BLOCKS.values())
def test_large_block(cls):
    data = token_bytes(1_048_576)

    block = cls(password=b'password', nonce=token_bytes(16), data=data)

    block.encrypt()
    assert block.data != data

    block.decrypt()
    assert block.data == data


@pytest.mark.parametrize("cls", AVAILABLE_BLOCKS.values())
@given(data=strategies.binary(), password=strategies.text())
def test_block_with_random_binary_data(cls, data, password):

    block = cls(password.encode('utf-8'), token_bytes(16), data)
    block.encrypt()

    assert block.data != data

    block.decrypt()
    assert block.data == data


@pytest.mark.parametrize("cls", AVAILABLE_BLOCKS.values())
def test_empty_block(cls):
    block = cls(b'password', token_bytes(16), b'')
    block.encrypt()

    assert block.data != b''

    block.decrypt()
    assert block.data == b''


@pytest.mark.parametrize("cls", AVAILABLE_BLOCKS.values())
def test_flip_bit(cls):
    block = cls(b'password', token_bytes(16), b'foobar')
    block.encrypt()

    data = bytearray(block.data)

    bit = random.randrange(0, len(data))
    data[bit] = data[bit] ^ 1

    block.data = bytes(data)

    with pytest.raises(Exception) as e:
        block.decrypt()

    assert 'InvalidTag' in str(e) or 'IntegrityError' in str(e)



import pytest

from spare import utils
from spare.errors import FileChangedDuringReadError


def test_abort_if_file_changed(temporary_path, loghandler):
    path = temporary_path / 'foo'
    path.touch()

    with pytest.raises(FileChangedDuringReadError):
        with utils.abort_if_file_changes_during_read(path):
            with path.open('w') as f:
                f.write('foobar')



from contextlib import contextmanager
from pathlib import Path
from spare.errors import FileChangedDuringReadError


def read_in_chunks(fileobj, chunksize):
    fileobj.seek(0)

    while True:
        chunk = fileobj.read(chunksize)

        if chunk:
            yield chunk
        else:
            break


@contextmanager
def abort_if_file_changes_during_read(path):
    before = Path(path).stat()
    yield
    after = Path(path).stat()

    for attr in ('st_mtime_ns', 'st_size', 'st_ino', 'st_dev'):
        if getattr(before, attr) != getattr(after, attr):
            raise FileChangedDuringReadError(path)



import grp
import hashlib
import io
import os
import os.path
import platform
import pwd
import stat

from cached_property import cached_property
from collections import defaultdict
from contextlib import suppress
from pathlib import Path
from spare import log
from spare.utils import abort_if_file_changes_during_read, read_in_chunks


def hash_implementation():
    """ We use blake2b with a 32 bytes digest size, which seems to be a good
    compromise between security, performance and short digest size.

    The checksum is not relevant for security, it is simply used to detect
    differences between files.

    """
    return hashlib.blake2b(digest_size=32)


def file_checksum(path):
    """ Generates the checksum for the given path. """

    with abort_if_file_changes_during_read(path):
        m = hash_implementation()

        with open(path, 'rb') as f:
            for chunk in read_in_chunks(f, io.DEFAULT_BUFFER_SIZE):
                m.update(chunk)

        return m.hexdigest()


def scandir(path, recurse=True, follow_symlinks=False):
    """ Runs os.scandir recursively with a coroutine interface.

    The coroutine interface is used to add extra paths to be scanned during
    the execution of the generator.

    """

    unscanned = {path}

    while unscanned:
        with os.scandir(unscanned.pop()) as iterator:
            for e in iterator:
                extra_path = yield e

                if extra_path:
                    unscanned.add(extra_path)

                if recurse and e.is_dir(follow_symlinks=follow_symlinks):
                    unscanned.add(e.path)


class Inventory(object):
    """ Manages metadata related to a local path.

    Every time the `scan` method is called, two dictionaries are created
    by walking the given path.

    The structure dictionary contains all the paths and metadata to the files
    found in the path. For example:

        {
            'pictures': {
                'type': 'directory',
                'user': 'denis',
                'group': 'staff',
                'mode': 41453,
                'size': 0,
                'mtime_ns': 1524208524622212344
            }
            'pictures/001.jpg': {
                'type': 'file',
                'user': 'denis',
                'group': 'staff',
                'mode': 41380,
                'size': 1024,
                'mtime_ns': 1524208524622245644
            },
            'bilder': {
                'type': 'symlink',
                'target': 'pictures',
                'user': 'denis',
                'group': 'staff',
                'mode': 41453,
                'size': 0,
                'mtime_ns': 1524208524622278944
            }
        }

    The structure is used to rebuild the directory tree when restoring a
    remote backup.

    The other dictionary, the data dictionary, contains the hashes of all
    files that we can upload remotely together with their paths. For example:

        {
            '0e5751c026e...': ['pictures/001.jpg']
        }

    As multiple files might have the same hash we can have multiple paths
    for a single hash. The data dictionary is used to figure out what we
    need to upload and what hash the file had at the time of the scan.

    It also acts as a basic way of deduplicate files. That is, files which
    have the same hash will be combined into a single upload and copied
    back when restoring.

    Note that due to the nature of our environment these scans cannot be relied
    on blindly. The structure changes and files are modified while they are
    being read or deleted before we get to process them further down the line.

    In this sense you should think of the inventory as an potentially
    inconsistent snapshot.

    """

    def __init__(self, path):
        self.path = Path(path)
        self.structure = {}
        self.files = defaultdict(list)

    @property
    def identity(self):
        return f'{platform.node()}:{self.path}'

    @cached_property
    def users(self):
        return {pw.pw_uid: pw.pw_name for pw in pwd.getpwall()}

    @cached_property
    def groups(self):
        return {gr.gr_gid: gr.gr_name for gr in grp.getgrall()}

    def cache_status(self, status):
        return {
            'user': self.users[status.st_uid],
            'group': self.groups[status.st_gid],
            'mode': status.st_mode,
            'size': status.st_size,
            'mtime_ns': status.st_mtime_ns,
            'inode': status.st_ino
        }

    def relative_path(self, path):
        return Path(path).relative_to(self.path)

    def absolute_path(self, path):
        return self.path / path

    def process_dir(self, path, status):
        self.structure[str(self.relative_path(path))] = {
            'type': 'directory',
            **self.cache_status(status)
        }

    def process_symlink(self, path, status, target):
        self.structure[str(self.relative_path(path))] = {
            'type': 'symlink',
            'target': str(self.relative_path(target)),
            **self.cache_status(status)
        }

    def process_file(self, path, status):
        relative_path = str(self.relative_path(path))
        is_empty = status.st_size == 0

        self.structure[relative_path] = {
            'type': 'file',
            'empty': is_empty,
            **self.cache_status(status)
        }

        if not is_empty:
            self.files[file_checksum(path)].append(relative_path)

    def clear(self):
        self.structure.clear()
        self.files.clear()

    def scan(self):
        self.clear()

        if self.path.is_dir():
            self.scan_directory(self.path)
        else:
            self.scan_entry(self.path)

    def scan_directory(self, path, recurse=True):
        scanner = scandir(path, recurse)
        extra = None

        with suppress(StopIteration):
            while True:
                entry = scanner.send(extra)
                extra = self.scan_entry(entry)

    def scan_entry(self, entry):
        if isinstance(entry, os.DirEntry):
            status = entry.stat(follow_symlinks=False)
        else:
            status = entry.stat()

        path = entry.__fspath__()
        scan_also = None

        if stat.S_ISCHR(status.st_mode):
            log.warn(f"Skipping character special device {path}")

        # we can't cover this on Travis currently as we don't have access to
        # a block device and can't create one without sudo
        elif stat.S_ISBLK(status.st_mode):  # pragma: no cover
            log.warn(f"Skipping block special device {path}")

        elif stat.S_ISFIFO(status.st_mode):
            log.warn(f"Skipping named pipe {path}")

        elif stat.S_ISSOCK(status.st_mode):
            log.warn(f"Skipping socket {path}")

        elif entry.is_symlink():
            target = Path(os.readlink(path))

            if not target.exists():
                log.warn(f"Skipping broken symlink {path}")

            elif self.path not in target.parents:
                if target.is_dir():
                    log.warn(f"Processing symlink {path} as a directory")
                    self.process_dir(path, status)

                    scan_also = path
                else:
                    log.warn(f"Processing symlink {path} as a file")
                    self.process_file(path, status)
            else:
                self.process_symlink(path, status, target)

        elif entry.is_dir():
            self.process_dir(path, status)

        elif entry.is_file():
            self.process_file(path, status)

        else:
            raise NotImplementedError  # pragma: no cover

        return scan_also



class SpareError(Exception):
    pass


class FileChangedDuringReadError(SpareError):
    """ Raised when a file has changed while it was read. """

    def __init__(self, path):
        self.path = path


class FileChangedBeforeUploadError(SpareError):
    """ Raised when a file changed after it was inventoried, but before
    it was uploaded.

    """

    def __init__(self, path):
        self.path = path


class PruneToZeroError(SpareError):
    """ Raised when trying to prune a bucket in a way that would not leave
    any snapshots behind.

    """

    def __init__(self, snapshots_left):
        self.snapshots_left = snapshots_left


class InvalidPrefixError(SpareError):
    """ Raised when a given prefix is invalid. """

    def __init__(self, prefix):
        self.prefix = prefix


class ExistingPrefixError(SpareError):
    """ Raised when a given prefix exists already. """

    def __init__(self, prefix):
        self.prefix = prefix


class BucketAlreadyLockedError(SpareError):
    """ Raised when bucket is already locked. """

    def __init__(self, bucket):
        self.bucket = bucket


class TargetPathNotEmpty(SpareError):
    """ Raised when a restore target path is not empty. """

    def __init__(self, path):
        self.path = path


class BucketNotLockedError(SpareError):
    """ Raised when a bucket was expected to be locked, but was not. """

    def __init__(self, bucket):
        self.bucket = bucket


class SnapshotMismatchError(SpareError):
    """ Raised when snapshot is attempted on a bucket which has snapshots
    from a different host/path on it.

    """

    def __init__(self, expected, found):
        self.expected = expected
        self.found = found


class BucketOtherwiseUsedError(SpareError):
    """ Raised when the bucket is not managed by spare. """

    def __init__(self, bucket):
        self.bucket = bucket



import pytest

from io import BytesIO
from spare.errors import FileChangedBeforeUploadError
from spare.errors import SnapshotMismatchError
from spare.errors import PruneToZeroError
from spare.inventory import Inventory
from spare.snapshot import SnapshotCollection


def test_empty_bucket(envoy):
    collection = SnapshotCollection(envoy)
    collection.load()

    assert not collection.snapshots


def test_save(envoy):
    collection = SnapshotCollection(envoy)

    s1 = collection.create()
    s1.save()

    collection.load()
    assert len(collection.snapshots) == 1

    s2 = collection.create()
    s2.save()

    collection.load()
    assert len(collection.snapshots) == 2

    s1.meta['foo'] = 'bar'
    s1.save()

    collection.load()
    assert collection.snapshots[0].meta['foo'] == 'bar'
    assert not collection.snapshots[1].meta


def test_prune(envoy):
    collection = SnapshotCollection(envoy)

    for i in range(0, 3):
        snapshot = collection.create()
        snapshot.meta['id'] = i
        snapshot.save()

    collection.load()
    assert len(collection.snapshots) == 3

    assert collection.snapshots[0].meta['id'] == 0
    assert collection.snapshots[1].meta['id'] == 1
    assert collection.snapshots[2].meta['id'] == 2

    collection.prune(keep=4)
    collection.load()

    assert len(collection.snapshots) == 3

    collection.prune(keep=3)
    collection.load()

    assert len(collection.snapshots) == 3

    collection.prune(keep=2)
    collection.load()

    assert len(collection.snapshots) == 2
    assert collection.snapshots[0].meta['id'] == 1
    assert collection.snapshots[1].meta['id'] == 2

    collection.prune(keep=1)
    collection.load()

    assert len(collection.snapshots) == 1
    assert collection.snapshots[0].meta['id'] == 2

    collection.prune(keep=1)
    collection.load()

    assert len(collection.snapshots) == 1
    assert collection.snapshots[0].meta['id'] == 2

    with pytest.raises(PruneToZeroError):
        collection.prune(keep=0)

    # make sure the internal files are not touched
    assert envoy.locked


def test_change_before_upload(envoy, temporary_path):
    path = temporary_path / 'foo'

    with path.open('w') as f:
        f.write('foo')

    inventory = Inventory(temporary_path)
    inventory.scan()

    with path.open('w') as f:
        f.write('bar')

    with pytest.raises(FileChangedBeforeUploadError):
        snapshot = SnapshotCollection(envoy).create()
        snapshot.backup(inventory)


def test_validate(envoy, temporary_path, loghandler):
    with (temporary_path / 'foo').open('w') as f:
        f.write('foo')

    with (temporary_path / 'bar').open('w') as f:
        f.write('bar')

    inventory = Inventory(temporary_path)
    inventory.scan()

    snapshot = SnapshotCollection(envoy).create()
    snapshot.backup(inventory)

    assert snapshot.validate()

    a, b = snapshot.meta['files']

    # remove one file
    envoy.delete(a)

    # replace another
    envoy.delete(b)
    envoy.send(b, BytesIO(b'baz'))

    # remove something from the structure
    snapshot.meta['structure'].popitem()
    snapshot.save()

    assert not snapshot.validate()

    records = [r for r in loghandler.records if r.level_name == 'ERROR']
    assert len(records) == 3

    log = '\n'.join(l.message for l in records)
    assert 'but got' in log
    assert 'is missing' in log
    assert 'is unknown' in log


def test_owner_mismatch(envoy, temporary_path):
    (temporary_path / 'foo').mkdir()
    (temporary_path / 'bar').mkdir()

    foo = Inventory(temporary_path / 'foo')
    foo.scan()

    bar = Inventory(temporary_path / 'bar')
    bar.scan()

    snapshot = SnapshotCollection(envoy).create()
    snapshot.backup(foo)

    with pytest.raises(SnapshotMismatchError):
        snapshot.backup(bar)

    with pytest.raises(SnapshotMismatchError):
        snapshot = SnapshotCollection(envoy).create()
        snapshot.backup(bar)



import os
import pytest
import socket
import tempfile

from contextlib import suppress
from pathlib import Path
from spare.inventory import Inventory


def test_empty_inventory(temporary_path):
    inventory = Inventory(temporary_path)
    inventory.scan()

    assert not inventory.structure
    assert not inventory.files


def test_single_file_inventory(temporary_path):
    with (temporary_path / 'foo').open('w') as f:
        f.write('foo')

    inventory = Inventory((temporary_path / 'foo'))
    inventory.scan()

    assert len(inventory.structure) == 1
    assert len(inventory.files) == 1

    assert '.' in inventory.structure
    assert list(inventory.files.values())[0][0] == '.'


def test_empty_file_inventory(temporary_path):
    (temporary_path / 'foo').touch()

    inventory = Inventory(temporary_path)
    inventory.scan()

    assert len(inventory.structure) == 1
    assert len(inventory.files) == 0

    assert inventory.structure['foo']['type'] == 'file'
    assert inventory.structure['foo']['empty'] is True


def test_single_symlink_inventory(temporary_path):
    (temporary_path / 'foo').touch()
    (temporary_path / 'bar').symlink_to((temporary_path / 'foo'))

    inventory = Inventory((temporary_path / 'bar'))
    inventory.scan()

    assert inventory.structure['.']['type'] == 'file'


def test_nested_paths(temporary_path):
    (temporary_path / 'foo').mkdir()

    with (temporary_path / 'foo' / 'bar').open('w') as f:
        f.write('foo')

    with (temporary_path / 'bar').open('w') as f:
        f.write('foo')

    inventory = Inventory(temporary_path)
    inventory.scan()

    assert 'foo' in inventory.structure
    assert 'foo/bar' in inventory.structure
    assert 'bar' in inventory.structure

    # two files with the same data
    assert len(inventory.files) == 1
    assert len(tuple(inventory.files.values())[0]) == 2


def test_symlinks(temporary_path):
    with (temporary_path / 'foo').open('w') as f:
        f.write('foo')

    (temporary_path / 'bar').symlink_to((temporary_path / 'foo'))

    inventory = Inventory(temporary_path)
    inventory.scan()

    assert len(inventory.files) == 1
    assert len(inventory.structure) == 2

    assert inventory.structure['bar']['target'] == 'foo'


def test_ignore_devices():
    # if we can scan /dev without errors we handle all the special devices
    inventory = Inventory('/dev')

    # try a bunch of folders to really get all kinds of devices
    # no need to do this recursively (which might not be a great idea)
    for path in ('/dev', '/dev/block', '/dev/disk/by-uuid'):
        with suppress(FileNotFoundError):
            inventory.scan_directory(Path(path), recurse=False)


def test_read_outside_symlink(temporary_path):

    with tempfile.TemporaryDirectory() as d:
        (Path(d) / 'outside').touch()
        (temporary_path / 'inside').symlink_to(Path(d) / 'outside')

        inventory = Inventory(temporary_path)
        inventory.scan()

        assert inventory.structure['inside']['type'] == 'file'


def test_read_outside_symlink_to_directory():

    with tempfile.TemporaryDirectory() as planet:
        with (Path(planet) / 'earth').open('w') as f:
            f.write('earth')

        with tempfile.TemporaryDirectory() as continent:
            with (Path(continent) / 'europe').open('w') as f:
                f.write('europe')

            (Path(continent) / 'planet').symlink_to(planet)

            with tempfile.TemporaryDirectory() as country:
                with (Path(country) / 'switzerland').open('w') as f:
                    f.write('switzerland')

                (Path(country) / 'continent').symlink_to(continent)

                inventory = Inventory(country)
                inventory.scan()

    assert inventory.structure['switzerland']['type'] == 'file'
    assert inventory.structure['continent']['type'] == 'directory'
    assert inventory.structure['continent/europe']['type'] == 'file'
    assert inventory.structure['continent/planet']['type'] == 'directory'
    assert inventory.structure['continent/planet/earth']['type'] == 'file'

    values = {p for paths in inventory.files.values() for p in paths}

    assert values == {
        'switzerland',
        'continent/europe',
        'continent/planet/earth',
    }


def test_detect_infinite_symlink_loop():

    with tempfile.TemporaryDirectory() as outside:
        with tempfile.TemporaryDirectory() as inside:
            (Path(outside) / 'inside').symlink_to(inside)
            (Path(inside) / 'outside').symlink_to(outside)

            with pytest.raises(OSError):
                inventory = Inventory(inside)
                inventory.scan()


def test_ignore_broken_symlink(temporary_path):
    (temporary_path / 'foo').touch()
    (temporary_path / 'bar').symlink_to((temporary_path / 'foo'))
    (temporary_path / 'foo').unlink()

    inventory = Inventory(temporary_path)
    inventory.scan()

    assert not inventory.structure


def test_ignore_fifo(temporary_path):
    fifo = (temporary_path / 'fifo')

    os.mkfifo(fifo)

    inventory = Inventory(temporary_path)
    inventory.scan()

    assert not inventory.structure
    assert not inventory.files


def test_ignore_socket_file(temporary_path):
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.bind(str(temporary_path / 'socket'))

    inventory = Inventory(temporary_path)
    inventory.scan()

    assert not inventory.structure
    assert not inventory.files



import os
import port_for
import pytest
import secrets
import sys
import tempfile

from boto3 import resource
from mirakuru import HTTPExecutor
from pathlib import Path
from spare.envoy import Envoy
from urllib.request import urlretrieve


MINIO_RELEASES = (
    ('darwin', 'https://dl.minio.io/server/minio/release/darwin-amd64/minio'),
    ('linux', 'https://dl.minio.io/server/minio/release/linux-amd64/minio')
)


@pytest.fixture(scope='session')
def access_key():
    return secrets.token_hex(20)


@pytest.fixture(scope='session')
def secret_key():
    return secrets.token_hex(40)


@pytest.fixture(scope='session')
def endpoint(minio):
    return minio.url


@pytest.fixture(scope='function')
def s3(access_key, secret_key, endpoint, minio_path):
    s3 = resource(
        service_name='s3',
        use_ssl=False,
        endpoint_url=endpoint.geturl(),
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key)

    yield s3

    for bucket in s3.buckets.iterator():
        bucket.objects.all().delete()
        bucket.delete()


@pytest.fixture(scope='session')
def minio_release():
    for key, url in MINIO_RELEASES:
        if key in sys.platform:
            return url

    raise RuntimeError(f"{sys.platform} is currently not supported")


@pytest.fixture(scope='session')
def minio_binary(request, minio_release):
    path = Path(request.config.cache.makedir('minio'))
    binary = path / 'minio'

    if not binary.exists():
        urlretrieve(minio_release, binary)
        binary.chmod(0o755)

    return binary


@pytest.fixture(scope='session')
def minio_path():
    with tempfile.TemporaryDirectory() as d:
        yield Path(d)


@pytest.fixture(scope='session')
def minio_port():
    return port_for.select_random()


@pytest.fixture(scope='session')
def minio(minio_binary, minio_path, minio_port, access_key, secret_key):
    os.environ['MINIO_ACCESS_KEY'] = access_key
    os.environ['MINIO_SECRET_KEY'] = secret_key
    os.environ['MINIO_UPDATE'] = 'off'

    address = f'127.0.0.1:{minio_port}'
    command = f'{minio_binary} server --address {address} {minio_path}'
    endpoint = f'http://{address}'

    # for some reason minio gets stuck at times if it prints to stdout
    # or if it is run as a non-shell command
    command += ' > /dev/null'

    executor = HTTPExecutor(command, url=endpoint, status=405, shell=True)
    executor.start()

    yield executor

    executor.stop()


@pytest.fixture(scope='function')
def temporary_path():
    with tempfile.TemporaryDirectory() as d:
        yield Path(d)


@pytest.fixture(scope='function')
def envoy(s3):
    with Envoy(s3, 'bucket', 'password') as envoy:
        yield envoy



from spare import log
from spare.envoy import Envoy
from spare.inventory import Inventory
from spare.recovery import Recovery
from spare.snapshot import SnapshotCollection


def create(path, s3, bucket, password, keep=1):
    inventory = Inventory(path)
    inventory.scan()

    with Envoy(s3, bucket, password) as envoy:

        collection = SnapshotCollection(envoy)
        collection.load()

        snapshot = collection.create()
        snapshot.backup(inventory)

        collection.prune(keep=keep)


def restore(path, s3, bucket, password, snapshot='latest'):
    with Envoy(s3, bucket, password) as envoy:

        collection = SnapshotCollection(envoy)
        collection.load()

        snapshot = collection.get(snapshot)

        recovery = Recovery(envoy, snapshot)
        recovery.restore(target=path)


def validate(s3, bucket, password):
    with Envoy(s3, bucket, password) as envoy:

        collection = SnapshotCollection(envoy)
        collection.load()

        valid = True

        for snapshot in collection.snapshots:
            if snapshot.validate():
                log.info(f"{snapshot.prefix} is valid")
            else:
                log.error(f"{snapshot.prefix} has errors")
                valid = False

        return valid



import click
import pdb
import sys
import traceback

from boto3 import resource
from logbook import StreamHandler
from spare.backup import create, restore, validate


VALID_PATH = click.Path(exists=True, file_okay=False)


def enable_post_mortem_debugging():  # pragma: no cover

    def hook(type, value, tb):
        if hasattr(sys, 'ps1') or not sys.stderr.isatty():
            sys.__excepthook__(type, value, tb)
        else:
            traceback.print_exception(type, value, tb)
            pdb.post_mortem(tb)

    sys.excepthook = hook


def s3_client(endpoint, access_key, secret_key):
    endpoint = '://' in endpoint and endpoint or f'https://{endpoint}'

    return resource(
        service_name='s3',
        endpoint_url=endpoint,
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key)


@click.group()
@click.option('--pdb', help="Enable post-mortem debugging", is_flag=True)
@click.option('--verbose', help="Print log messages to stdout", is_flag=True)
@click.pass_context
def cli(ctx, pdb, verbose):  # pragma: no cover
    if pdb:
        enable_post_mortem_debugging()

    level = verbose and 'INFO' or 'WARNING'
    StreamHandler(sys.stdout, level=level).push_application()


@cli.command(name='create')
@click.option('--endpoint', envvar='SPARE_ENDPOINT', required=True)
@click.option('--access-key', envvar='SPARE_ACCESS_KEY', required=True)
@click.option('--secret-key', envvar='SPARE_SECRET_KEY', required=True)
@click.option('--password', envvar='SPARE_PASSWORD', required=True)
@click.option('--bucket', envvar='SPARE_BUCKET', required=True)
@click.option('--path', envvar='SPARE_PATH', type=VALID_PATH, required=True)
def create_cli(endpoint, access_key, secret_key, path, password, bucket):
    s3 = s3_client(endpoint, access_key, secret_key)
    create(path, s3, bucket, password)


@cli.command(name='restore')
@click.option('--endpoint', envvar='SPARE_ENDPOINT', required=True)
@click.option('--access-key', envvar='SPARE_ACCESS_KEY', required=True)
@click.option('--secret-key', envvar='SPARE_SECRET_KEY', required=True)
@click.option('--password', envvar='SPARE_PASSWORD', required=True)
@click.option('--bucket', envvar='SPARE_BUCKET', required=True)
@click.option('--path', envvar='SPARE_PATH', type=VALID_PATH, required=True)
def restore_cli(endpoint, access_key, secret_key, path, password, bucket):
    s3 = s3_client(endpoint, access_key, secret_key)
    restore(path, s3, bucket, password)


@cli.command(name='validate')
@click.option('--endpoint', envvar='SPARE_ENDPOINT', required=True)
@click.option('--access-key', envvar='SPARE_ACCESS_KEY', required=True)
@click.option('--secret-key', envvar='SPARE_SECRET_KEY', required=True)
@click.option('--password', envvar='SPARE_PASSWORD', required=True)
@click.option('--bucket', envvar='SPARE_BUCKET', required=True)
def validate_cli(endpoint, access_key, secret_key, password, bucket):
    s3 = s3_client(endpoint, access_key, secret_key)
    rc = 0 if validate(s3, bucket, password) else 1

    sys.exit(rc)



import json
import ulid

from io import BytesIO
from spare import log
from spare.errors import FileChangedBeforeUploadError
from spare.errors import PruneToZeroError
from spare.errors import SnapshotMismatchError
from spare.inventory import hash_implementation
from spare.utils import abort_if_file_changes_during_read


class SnapshotCollection(object):
    """ Manages the snapshots stored on a bucket.

    Snapshots are created using 'ulid' to automatically keep them in order
    of creation (from oldest to newest).

    Snapshots store the structure of the inventory they backed up as well
    as all the file-digests they are linked with. This ends up working
    a bit like reference counting - as long as a snapshot links to an
    uploaded file, that file is kept around.

    When snapshots are deleted/pruned, all the files which are no longer
    referenced by any snapshot get deleted.

    """

    def __init__(self, envoy):
        self.envoy = envoy
        self.clear()

    def clear(self):
        self.snapshots = []

    def load(self):
        """ Load all snapshots from the bucket. """

        self.clear()

        for prefix in sorted(self.envoy.prefixes('snapshot')):
            self.snapshots.append(Snapshot.load(self.envoy, prefix))

    def create(self):
        return Snapshot(self.envoy)

    def get(self, prefix='latest'):
        """ Gets the given snapshot by prefix or the latest one (default). """

        if prefix == 'latest':
            return self.snapshots[-1]

        return next((s for s in self.snapshots if s.prefix == prefix), None)

    def prune(self, keep=1):
        """ Prunes the snapshots and all the files associated with them. """

        # make sure we are up to date
        self.load()

        if keep < 1:
            raise PruneToZeroError(len(self.snapshots))

        if len(self.snapshots) <= keep:
            return

        # delete the meatdata of old snapshots
        old, self.snapshots = self.snapshots[:-keep], self.snapshots[-keep:]

        for snapshot in old:
            snapshot.delete()

        # delete the prefixes we don't recognise
        known = set()

        for snapshot in self.snapshots:
            known.add(snapshot.prefix)
            known.update(d for d in snapshot.meta.get('files', ()))

        for prefix in self.envoy.prefixes():
            if prefix not in known:
                self.envoy.delete(prefix)


class Snapshot(object):
    """ A single snapshot, responsible for backing up inventories. """

    def __init__(self, envoy, prefix=None, meta=None):
        self.envoy = envoy
        self.prefix = prefix or f'snapshot_{ulid.new()}'
        self.meta = meta or {}

    @classmethod
    def load(cls, envoy, prefix):
        with BytesIO() as buffer:
            envoy.recv(prefix, buffer)
            meta = json.loads(buffer.getvalue().decode('utf-8'))

        return cls(envoy, prefix, meta)

    @property
    def neighbours(self):
        collection = SnapshotCollection(self.envoy)
        collection.load()

        for snapshot in collection.snapshots:
            if snapshot.prefix != self.prefix:
                yield snapshot

    def ensure_identity_match(self, inventory):
        """ Each inventory we backup has an identity associated with it
        (hostname + path). When creating a new snapshot we ensure that
        this identity matches, because we want each hostname/path combination
        to be stored in a separate bucket.

        """
        if 'identity' in self.meta:
            if self.meta['identity'] != inventory.identity:
                raise SnapshotMismatchError(
                    expected=inventory.identity,
                    found=self.meta['identity']
                )

        for snapshot in self.neighbours:
            if snapshot.meta['identity'] != inventory.identity:
                raise SnapshotMismatchError(
                    expected=inventory.identity,
                    found=snapshot.meta['identity']
                )

    def backup(self, inventory):
        """ Backup the given inventory. """

        log.info(f"Backing up {inventory.path}")

        self.ensure_identity_match(inventory)
        uploaded = set(self.envoy.prefixes())

        for digest, paths in inventory.files.items():
            if digest in uploaded:
                continue

            for path in paths:
                log.info(f"Uploading {path}")

            path = inventory.absolute_path(paths[0])
            m = hash_implementation()

            with abort_if_file_changes_during_read(path):
                with open(path, 'rb') as f:
                    self.envoy.send(digest, f, before_encrypt=m.update)

            if digest != m.hexdigest():
                raise FileChangedBeforeUploadError(path)

        self.meta['files'] = inventory.files
        self.meta['structure'] = inventory.structure
        self.meta['identity'] = inventory.identity

        self.save()

        log.info(f"Completed {self.prefix}")

    def save(self):
        """ Save the snapshot metadata to the bucket. """

        if self.envoy.is_known_prefix(self.prefix):
            self.envoy.delete(self.prefix)

        with BytesIO() as buffer:
            buffer.write(json.dumps(self.meta).encode('utf-8'))
            self.envoy.send(self.prefix, buffer)

    def delete(self):
        """ Deletes the snapshot data, but *not* the associated data. The
        latter is done by the `prune` call in the `SnapshotCollection` class.

        """
        self.envoy.delete(self.prefix)

    def validate(self):
        """ Validates the snapshot data by downloading and validating the
        checksum of all the files in the envoy.

        Returns true if the snapshot is valid. Errors are added to the log.

        """

        class NullBuffer(object):
            def write(self, bytes):
                pass

        prefixes = set(self.envoy.prefixes())

        null = NullBuffer()
        success = True

        def fail(message):
            nonlocal success
            success = False
            log.error(message)

        for digest in self.meta.get('files', ()):

            if digest not in prefixes:
                fail(f"{digest} is unknown")

            else:
                m = hash_implementation()
                self.envoy.recv(digest, null, after_decrypt=m.update)

                if digest != m.hexdigest():
                    fail(f"Expected {digest} but got {m.hexdigest()}")

            for path in self.meta['files'][digest]:
                if path not in self.meta['structure']:
                    fail(f"the metadata for {path} is missing")

        return success



import re
import secrets

from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from spare.block import DEFAULT_BLOCK, AVAILABLE_BLOCKS
from spare.errors import BucketAlreadyLockedError
from spare.errors import BucketNotLockedError
from spare.errors import BucketOtherwiseUsedError
from spare.errors import ExistingPrefixError
from spare.errors import InvalidPrefixError
from spare.utils import read_in_chunks


class Envoy(object):
    """ Provides the bridge between an unencrypted local file and a remote file
    encrypted in chunks.

    The keys used for each chunks are of the following format:

        prefix/000000001-nonce

    For example:

        my-file/000000001-c3f543e56704af2ca4779a7d530836cc
        my-file/000000002-4489c3d7ff0e090ad1a1260efa2f5084

    As the number indicates, we are limited to 999'999'999 blocks of 1 MiB
    each, so the largest file we can accept after compression and encryption
    is 953.674 Tebibytes.

    """

    # valid prefixes are limited to a-z A-Z 0-9 and underscore. It may not
    # begin with a dot (reserved for envoy-internal files) and it must
    # be at least two characters long
    valid_prefix = re.compile(r'^[a-zA-Z0-9_]{1}[a-zA-Z0-9_\.]+$')

    def __init__(self, s3, bucket, password, block=DEFAULT_BLOCK):
        self.s3 = s3
        self.bucket = s3.Bucket(bucket)
        self.bucket_name = bucket
        self.password = password
        self.block_class = AVAILABLE_BLOCKS[DEFAULT_BLOCK]

        # you should not change these values outside of unit tests!
        self.blocksize = 1_048_576  # found to be good value through testing
        self.noncesize = 16

    def __enter__(self):
        self.lock()
        return self

    def __exit__(self, *args):
        self.unlock()

    def lock(self):
        self.ensure_bucket_exists()
        self.ensure_bucket_is_ours()

        if self.locked:
            raise BucketAlreadyLockedError(self.bucket_name)

        with BytesIO() as f:
            self.bucket.upload_fileobj(f, '.lock')

    @property
    def locked(self):
        return self.is_known_prefix('.lock')

    def unlock(self):
        if self.is_known_prefix('.lock'):
            self.bucket.objects.filter(Prefix='.lock').delete()

    def ensure_locked(self):
        if not self.locked:
            raise BucketNotLockedError(self.bucket_name)

    def ensure_bucket_exists(self):
        if not self.bucket.creation_date:
            self.bucket.create()

            with BytesIO(b'https://github.com/seantis/spare') as f:
                self.bucket.upload_fileobj(f, '.spare')

    def ensure_bucket_is_ours(self):
        if not self.is_known_prefix('.spare'):
            raise BucketOtherwiseUsedError(self.bucket_name)

    def ensure_prefix_unknown(self, prefix):
        if self.is_known_prefix(prefix):
            raise ExistingPrefixError(prefix)

    def ensure_valid_prefix(self, prefix):
        if not prefix or not self.valid_prefix.match(prefix):
            raise InvalidPrefixError(prefix)

    def spawn_block(self, nonce, data):
        return self.block_class(
            password=self.password.encode('utf-8'),
            nonce=nonce.encode('utf-8'),
            data=data
        )

    def generate_nonce(self):
        return secrets.token_hex(self.noncesize)

    def extract_nonce(self, key):
        return key.split('-')[-1]

    def extract_prefix(self, key):
        return key.split('/')[0]

    def is_first_block(self, key):
        return '1-' in key

    def is_known_prefix(self, prefix):
        for obj in self.bucket.objects.filter(Prefix=prefix, MaxKeys=1):
            return True

        return False

    def keys(self, prefix=None):
        for obj in self.bucket.objects.filter(Prefix=prefix or ''):
            if not obj.key.startswith('.'):
                yield obj.key

    def prefixes(self, prefix=None):
        for key in self.keys(prefix=prefix):
            if self.is_first_block(key):
                yield self.extract_prefix(key)

    def delete(self, prefix):
        self.ensure_locked()
        self.ensure_valid_prefix(prefix)
        self.bucket.objects.filter(Prefix=prefix).delete()

    def send(self, prefix, fileobj, before_encrypt=None):
        self.ensure_locked()
        self.ensure_valid_prefix(prefix)
        self.ensure_prefix_unknown(prefix)

        chunks = read_in_chunks(fileobj, self.blocksize)

        # uploading chunks in threads makes little difference in memory and cpu
        # usage, but it does speed up large downloads by 20%
        #
        # further up the stack threads are not much more effective, but they
        # use much more cpu/memory
        #
        # maybe having some kind of pipeline of all chunks (not just per file
        # would improve things further, but for now this seems reasonable
        # enough)
        def upload_block(name, block):
            with BytesIO(block.data) as buffer:
                self.bucket.upload_fileobj(buffer, name)

        with ThreadPoolExecutor() as executor:
            for n, chunk in enumerate(chunks, start=1):
                nonce = self.generate_nonce()
                block = self.spawn_block(nonce, chunk)

                if before_encrypt is not None:
                    before_encrypt(chunk)

                block.encrypt()
                block_name = f'{prefix}/{n:0>9}-{nonce}'

                executor.submit(upload_block, block_name, block)

    def recv(self, prefix, fileobj, after_decrypt=None):
        self.ensure_valid_prefix(prefix)

        for obj in self.bucket.objects.filter(Prefix=prefix):
            nonce = self.extract_nonce(obj.key)

            with BytesIO() as buffer:
                self.bucket.download_fileobj(obj.key, buffer)

                buffer.seek(0)
                block = self.spawn_block(nonce, buffer.read())

            block.decrypt()

            if after_decrypt is not None:
                after_decrypt(block.data)

            fileobj.write(block.data)



from io import BytesIO
from spare.download import Download, DownloadManager


def test_digestless_download(envoy, temporary_path):
    envoy.send('foo', BytesIO(b'foo'))

    with DownloadManager(envoy) as manager:
        download = Download('foo', 'wrong-digest')
        download.to(temporary_path / 'foo')

        manager.queue(download)

    assert (temporary_path / 'foo').open('rb').read() == b'foo'


def test_wrong_digest_download(envoy, temporary_path, loghandler):
    envoy.send('foo', BytesIO(b'foo'))

    with DownloadManager(envoy) as manager:
        download = Download('foo', 'wrong-digest')
        download.to(temporary_path / 'foo')

        manager.queue(download)

    # the download will still work
    assert (temporary_path / 'foo').open('rb').read() == b'foo'

    # but we get an error int he logs
    assert "Unexpected checksum for" in loghandler.records[0].message



def test_access(s3):
    assert not tuple(s3.buckets.iterator())
    s3.create_bucket(Bucket='foobar')
    assert tuple(s3.buckets.iterator())


def test_buckets_cleared_between_tests(s3):
    assert not tuple(s3.buckets.iterator())



import os
import secrets
import pytest

from pathlib import Path
from spare.backup import create, restore, validate
from spare.envoy import Envoy
from spare.errors import TargetPathNotEmpty
from spare.snapshot import SnapshotCollection
from tempfile import TemporaryDirectory
from unittest import mock


def test_create(s3, temporary_path):
    envoy = Envoy(s3, 'my-bucket', 'password')

    with (temporary_path / 'foo').open('w') as f:
        f.write('foo')

    create(temporary_path, s3, 'my-bucket', 'password')
    prefixes = set(envoy.prefixes())
    assert len(prefixes) == 2

    create(temporary_path, s3, 'my-bucket', 'password')
    prefixes = set(envoy.prefixes())
    assert len(prefixes) == 2

    with (temporary_path / 'bar').open('w') as f:
        f.write('bar')

    create(temporary_path, s3, 'my-bucket', 'password')
    prefixes = set(envoy.prefixes())
    assert len(prefixes) == 3

    (temporary_path / 'bar').unlink()

    create(temporary_path, s3, 'my-bucket', 'password')
    prefixes = set(envoy.prefixes())
    assert len(prefixes) == 2


def test_large_file(s3):
    content = secrets.token_bytes(1024*1024)

    with TemporaryDirectory() as d:
        d = Path(d)

        with (d / 'foo').open('wb') as f:
            f.write(content)

        create(d, s3, 'my-bucket', 'password')

    with TemporaryDirectory() as d:
        d = Path(d)

        restore(d, s3, 'my-bucket', 'password')

        with (d / 'foo').open('rb') as f:
            assert f.read() == content


def test_restore(s3):

    with TemporaryDirectory() as d:
        d = Path(d)

        # normal files
        with (d / 'foo').open('w') as f:
            f.write('foo')

        with (d / 'bar').open('w') as f:
            f.write('bar')

        # symlinks
        (d / 'link-to-foo').symlink_to(d / 'foo')

        # hardlinks
        os.link(d / 'bar', d / 'hardlink-to-bar')

        # directories
        (d / 'dir').mkdir()
        (d / 'dir' / 'foo').mkdir()
        (d / 'dir' / 'foo' / 'bar').mkdir()

        # executables
        (d / 'exe').touch()
        (d / 'exe').chmod(0o755)

        create(d, s3, 'my-bucket', 'password')

    with TemporaryDirectory() as d:
        d = Path(d)

        restore(d, s3, 'my-bucket', 'password')

        # normal files
        assert (d / 'foo').open('r').read() == 'foo'
        assert (d / 'bar').open('r').read() == 'bar'

        # symlinks
        assert (d / 'link-to-foo').is_symlink()
        assert os.readlink(d / 'link-to-foo') == str(d / 'foo')
        assert (d / 'link-to-foo').open('r').read() == 'foo'

        # hardlinks
        assert (d / 'hardlink-to-bar').stat().st_ino\
            == (d / 'bar').stat().st_ino

        # directories
        assert (d / 'dir').is_dir()
        assert (d / 'dir' / 'foo').is_dir()
        assert (d / 'dir' / 'foo' / 'bar').is_dir()

        # executables
        assert (d / 'exe').stat().st_mode & 0o777 == 0o755


def test_restore_complex_hardlinks_case(s3):
    with TemporaryDirectory() as d:
        d = Path(d)

        for child in ('foo', 'bar'):
            (d / child).mkdir()

            with (d / child / 'a').open('wb') as f:
                f.write(b'foo')

            # a <- b <- c
            os.link((d / child / 'a'), (d / child / 'b'))
            os.link((d / child / 'b'), (d / child / 'c'))

            assert (d / child / 'a').stat().st_ino\
                == (d / child / 'b').stat().st_ino\
                == (d / child / 'c').stat().st_ino

        create(d, s3, 'my-bucket', 'password')

    with TemporaryDirectory() as d:
        d = Path(d)

        # we should see two hierarchies with three files sharing a single
        # inode for each directory (no overlap between directories)
        restore(d, s3, 'my-bucket', 'password')
        inodes = set()

        for child in ('foo', 'bar'):
            assert (d / child).is_dir()
            assert (d / child / 'a').open('rb').read() == b'foo'
            assert (d / child / 'b').open('rb').read() == b'foo'
            assert (d / child / 'c').open('rb').read() == b'foo'

            assert (d / child / 'a').stat().st_ino\
                == (d / child / 'b').stat().st_ino\
                == (d / child / 'c').stat().st_ino

            inode = (d / child / 'a').stat().st_ino

            assert inode not in inodes
            inodes.add(inode)

    # ensure that this setup only results in a single download
    with mock.patch('spare.download.DownloadManager.fetch') as m:
        with TemporaryDirectory() as d:
            restore(Path(d), s3, 'my-bucket', 'password')

    assert m.call_count == 1


def test_restore_previous_snapshot(s3):

    with TemporaryDirectory() as d:
        d = Path(d)

        with (d / 'foo').open('w') as f:
            f.write('foo')

        create(d, s3, 'my-bucket', 'password')

        collection = SnapshotCollection(Envoy(s3, 'my-bucket', 'password'))
        collection.load()

        assert len(collection.snapshots) == 1

        with (d / 'foo').open('w') as f:
            f.write('bar')

        (d / 'bar').mkdir()

        create(d, s3, 'my-bucket', 'password', keep=2)

        collection.load()
        assert len(collection.snapshots) == 2

    with TemporaryDirectory() as d:
        d = Path(d)

        restore(d, s3, 'my-bucket', 'password', collection.snapshots[0].prefix)

        assert (d / 'foo').open('r').read() == 'foo'
        assert not (d / 'bar').exists()

    with TemporaryDirectory() as d:
        d = Path(d)

        restore(d, s3, 'my-bucket', 'password')

        assert (d / 'foo').open('r').read() == 'bar'
        assert (d / 'bar').exists()


def test_restore_to_non_empty_directory(s3, temporary_path):
    (temporary_path / 'foo').touch()
    create(temporary_path, s3, 'my-bucket', 'password')

    with pytest.raises(TargetPathNotEmpty):
        restore(temporary_path, s3, 'my-bucket', 'password')


def test_validate(s3, temporary_path):
    with TemporaryDirectory() as d:
        d = Path(d)

        with (d / 'foo').open('w') as f:
            f.write('foo')

        with (d / 'foo').open('w') as f:
            f.write('bar')

        create(d, s3, 'my-bucket', 'password')

    assert validate(s3, 'my-bucket', 'password')

    with Envoy(s3, 'my-bucket', 'password') as envoy:
        envoy.delete(next(envoy.prefixes()))

    assert not validate(s3, 'my-bucket', 'password')



import os
import grp
import pwd

from cached_property import cached_property
from itertools import groupby
from pathlib import Path
from spare import log, FOLLOW_SYMLINKS
from spare.errors import TargetPathNotEmpty
from spare.download import Download, DownloadManager


class Recovery(object):
    """ Responsible for the recovery of snapshots to any given path.

    Whereas our backing up procedures try to be conservative with cpu and
    memory usage, this module tries to maximise recovery speed. The assumption
    is that recovery happens on a system that is otherwise unused.

    """

    def __init__(self, envoy, snapshot):
        self.envoy = envoy
        self.snapshot = snapshot

    @cached_property
    def uids(self):
        return {pw.pw_name: pw.pw_uid for pw in pwd.getpwall()}

    @cached_property
    def gids(self):
        return {gr.gr_name: gr.gr_gid for gr in grp.getgrall()}

    def restore(self, target):
        log.info(f"Restoring {target}")

        target = Path(target)
        target.mkdir(exist_ok=True)

        for path in target.iterdir():
            raise TargetPathNotEmpty(path)

        log.info(f"Restoring folder structure of {target}")
        self.restore_structure(target)

        log.info(f"Downloading data for {target}")
        self.download_data(target)

        log.info(f"Restored {target}")

    def restore_structure(self, target):
        structure = self.snapshot.meta['structure']

        for path in sorted(structure):
            status = structure[path]
            path = target / path

            if status['type'] == 'directory':
                path.mkdir()

            elif status['type'] == 'symlink':
                path.symlink_to(target / status['target'])

            elif status['type'] == 'file':
                path.touch()

            else:
                raise NotImplementedError  # pragma: no cover

            uid = self.uids[status['user']]
            gid = self.gids[status['group']]

            os.chmod(path, status['mode'], **FOLLOW_SYMLINKS)
            os.chown(path, uid, gid, **FOLLOW_SYMLINKS)

    def download_data(self, target):
        structure = self.snapshot.meta['structure']

        def inode(path):
            return structure[path]['inode']

        def by_inode(paths):
            return groupby(sorted(paths, key=inode), key=inode)

        with DownloadManager(self.envoy) as download_manager:
            for digest, paths in self.snapshot.meta['files'].items():
                download = Download(prefix=digest, digest=digest)

                log.info(f"Downloading {paths[0]}")

                for _, paths in by_inode(paths):
                    path, *rest = (target / p for p in paths)
                    download.to(path, hardlinks=rest)

                download_manager.queue(download)



import os
import shutil

from concurrent.futures import ThreadPoolExecutor
from spare import log, FOLLOW_SYMLINKS
from spare.inventory import hash_implementation


class Download(object):
    """ Represents a single download.

    Locally duplicated files are only stored once on the remote and we want
    to only download them once as well.

    Therefore we use a single prefix and n targets. Each target is a list of
    1-n paths. The first path is a copy of the original digest, all the other
    paths of the same target ar hardlinks to the copy of the original digest.

    For example, say we have this folder structure where all the files
    have the same content and the files in each folder have the same inode:

    / foo / a (inode 100)
    / foo / b (inode 100)
    / foo / c (inode 100)

    / bar / a (inode 200)
    / bar / b (inode 200)
    / bar / c (inode 200)

    This should result in a single download of this form:

    Download(prefix='xxx', targets=[
        ('/foo/a', '/foo/b', '/foo/c'),
        ('/bar/a', '/bar/b', '/bar/c')
    ])

    This in turn will lead to a download to /foo/a, hardlinks to /foo/a from
    /foo/b /foo/c). As well as a copy of /foo/a to /bar/a and hardlinks from
    /bar/b and /bar/c to /bar/a.

    """

    __slots__ = ('prefix', 'digest', 'targets')

    def __init__(self, prefix, digest):
        self.prefix = prefix
        self.digest = digest
        self.targets = []

    def to(self, path, hardlinks=()):
        self.targets.append((path, *hardlinks))


class DownloadManager(object):
    """ Takes download objects and downloads them using a threadpool executor.

    In addition the download targets are realised after the download (incl.
    copies and hardlinks).

    Since the download manager works with threads it should be used with
    a with clause:

        with DownloadManager(envoy) as manager:
            download = Download('prefix', 'digest')
            download.to(path, hardlinks)
            download.to(path, hardlinks)

            manager.queue(download)

    """

    def __init__(self, envoy):
        self.envoy = envoy
        self.executor = ThreadPoolExecutor()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.executor.shutdown(wait=True)

    def queue(self, download):
        self.executor.submit(self.fetch, download)

    def fetch(self, download):
        m = hash_implementation()
        genesis = download.targets[0][0]

        with open(genesis, 'wb') as f:
            self.envoy.recv(download.prefix, f, after_decrypt=m.update)

        # the following branch is actually covered, but coverage.py does not
        # consider it as such - I've tried all the tricks in the book and it
        # will still not capture it, so 'no cover' it is
        if download.digest != m.hexdigest():  # pragma: no cover
            paths = (p for paths in download.targets for p in paths)

            for path in paths:
                log.error((
                    f"Unexpected checksum for {path}, "
                    f"expected {download.digest}, "
                    f"got {m.hexdigest()}"
                ))

        for paths in download.targets:
            clone, *links = paths

            if clone != genesis:
                shutil.copyfile(genesis, clone, **FOLLOW_SYMLINKS)
                shutil.copystat(genesis, clone, **FOLLOW_SYMLINKS)

            for link in links:
                # files are touched during structure creation, which is a
                # problem for hard links
                os.unlink(link)

                os.link(clone, link, **FOLLOW_SYMLINKS)



import secrets
import pytest

from io import BytesIO
from spare.block import AVAILABLE_BLOCKS
from spare.envoy import Envoy
from spare.errors import BucketAlreadyLockedError
from spare.errors import BucketNotLockedError
from spare.errors import BucketOtherwiseUsedError
from spare.errors import ExistingPrefixError
from spare.errors import InvalidPrefixError


@pytest.mark.parametrize('block', AVAILABLE_BLOCKS)
def test_single_block(s3, block):
    with Envoy(s3, 'my-bucket', 'my-password', block) as envoy:
        envoy.send('document', BytesIO(b'secret'))

        blocks = tuple(envoy.bucket.objects.filter(Prefix='document'))

        assert len(blocks) == 1
        assert blocks[0].get()['Body'].read() != b'secret'

        output = BytesIO()
        envoy.recv('document', output)

        output.seek(0)
        assert output.read() == b'secret'


@pytest.mark.parametrize('block', AVAILABLE_BLOCKS)
def test_multiple_blocks(s3, block):
    with Envoy(s3, 'my-bucket', 'my-password', block) as envoy:
        envoy.blocksize = 1024

        data = secrets.token_bytes(1024 * 2)
        envoy.send('document', BytesIO(data))

        blocks = tuple(envoy.bucket.objects.filter(Prefix='document'))

        assert len(blocks) == 2

        output = BytesIO()
        envoy.recv('document', output)

        output.seek(0)
        assert output.read() == data


def test_replace(s3):
    with Envoy(s3, 'my-bucket', 'my-password') as envoy:
        envoy.blocksize = 1024

        initial = secrets.token_bytes(1024 * 2)
        replacement = secrets.token_bytes(1024 * 1)

        envoy.send('document', BytesIO(initial))
        assert sum(1 for k in envoy.keys()) == 2

        with pytest.raises(ExistingPrefixError):
            envoy.send('document', BytesIO(replacement))
            assert sum(1 for k in envoy.keys()) == 1

        envoy.delete('document')
        envoy.send('document', BytesIO(replacement))

        output = BytesIO()
        envoy.recv('document', output)

        output.seek(0)
        assert output.read() == replacement


def test_delete(s3):
    with Envoy(s3, 'my-bucket', 'my-password') as envoy:
        envoy.send('foo', BytesIO(b'foo'))
        envoy.send('bar', BytesIO(b'bar'))
        assert sum(1 for k in envoy.keys()) == 2

        envoy.delete('bar')
        assert sum(1 for k in envoy.keys()) == 1

        with BytesIO() as output:
            envoy.recv('foo', output)
            assert output.getvalue() == b'foo'


def test_invalid_prefix(s3):
    with Envoy(s3, 'bucket', 'password') as envoy:
        # no dashes
        with pytest.raises(InvalidPrefixError):
            envoy.send('no-dashes', BytesIO())

        # at least 2 characters long
        with pytest.raises(InvalidPrefixError):
            envoy.send('x', BytesIO())

        # no dots
        with pytest.raises(InvalidPrefixError):
            envoy.send('.hidden', BytesIO())


def test_prefixes(s3):
    with Envoy(s3, 'bucket', 'password') as envoy:
        envoy.blocksize = 1024

        envoy.ensure_bucket_exists()
        assert not next(envoy.prefixes(), None)

        data = secrets.token_bytes(1024 * 2)
        envoy.send('foo', BytesIO(data))

        assert sum(1 for k in envoy.keys()) == 2
        assert tuple(envoy.prefixes()) == ('foo', )


def test_lock(s3):
    with Envoy(s3, 'bucket', 'password'):
        pass

    with Envoy(s3, 'bucket', 'password'):
        pass

    with pytest.raises(BucketAlreadyLockedError):
        with Envoy(s3, 'bucket', 'password'):
            with Envoy(s3, 'bucket', 'password'):
                pass


def test_lock_enforced(s3):
    with Envoy(s3, 'bucket', 'password'):
        pass

    with pytest.raises(BucketNotLockedError):
        Envoy(s3, 'bucket', 'password').send('foo', BytesIO())


def test_envoy_fail_on_foreign_buckets(s3):
    bucket = Envoy(s3, 'bucket', 'password').bucket
    bucket.create()
    bucket.upload_fileobj(BytesIO(b'foo'), 'bar')

    with pytest.raises(BucketOtherwiseUsedError):
        with Envoy(s3, 'bucket', 'password'):
            pass



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

name = 'spare'
description = 'One S3 backup, encrypted on the fly.'
version = '0.1.0'


def get_long_description():
    readme = open('README.rst').read()
    history = open('HISTORY.rst').read()

    # cut the part before the description to avoid repetition on pypi
    readme = readme[readme.index(description) + len(description):]

    return '\n'.join((readme, history))


setup(
    name=name,
    version=version,
    description=description,
    long_description=get_long_description(),
    url='http://github.com/seantis/spare',
    author='Seantis GmbH',
    author_email='info@seantis.ch',
    license='MIT',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=name.split('.')[:-1],
    include_package_data=True,
    zip_safe=False,
    platforms='any',

    # we require 3.6.2 (not just 3.6.*), due to this bug:
    # https://bugs.python.org/issue29581
    python_requires='>=3.6.2',

    install_requires=[
        'boto3',
        'cached_property',
        'click',
        'cryptography',
        'logbook',
        'miscreant',
        'ulid-py',
    ],
    extras_require=dict(
        test=[
            'hypothesis',
            'mirakuru',
            'flake8',
            'port-for',
            'pytest',
            'pytest-cov',
            'pytest-flake8',
            'pytest-logbook'
        ],
    ),
    entry_points={
        'console_scripts': 'spare=spare.cli:cli'
    },
    classifiers=[
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Archiving :: Backup',
    ]
)



import lzma
import hashlib

from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from miscreant.aead import AEAD


AVAILABLE_BLOCKS = {}
DEFAULT_BLOCK = 'aes-siv'


class Block(object):
    """ Interface to implement a block. Each block is meant to encrypt and
    decrypt data in memory.

    Blocks may share a single passsword, but they are virtually guaranteed to
    have independent nonces (randomly chosen from a pool of 16^16 values).

    Blocks should also compress/decompress data, the used encryption should
    not use any padding and it should authenticate during/after decryption.

    """

    __slots__ = ('password', 'nonce', 'data')

    def __init_subclass__(cls, id, **kwargs):
        assert id not in AVAILABLE_BLOCKS
        AVAILABLE_BLOCKS[id] = cls

        super().__init_subclass__(**kwargs)

    def __init__(self, password, nonce, data):
        """ Called with the password, nonce and data to encrypt/decrypt.

        All parameters ought to be in bytes.
        """
        raise NotImplementedError  # pragma: nocover

    def encrypt(self):
        """ Encrypts self.data in-place, doesn't return anything. """

        raise NotImplementedError  # pragma: nocover

    def decrypt(self):
        """ Decrypts self.data in-place, doesn't return anything. """

        raise NotImplementedError  # pragma: nocover


class LZMA_AES_SIV_Block(Block, id='aes-siv'):
    """ The default block implementation, using AES-SIV by miscreant.

    Blocks are compressed using LZMA before being encrypted.

    """

    __slots__ = ('password', 'nonce', 'data')

    def __init__(self, password, nonce, data):
        self.password = hashlib.sha512(password).digest()
        self.nonce = nonce
        self.data = data

    @property
    def aead(self):
        return AEAD('AES-SIV', self.password)

    def encrypt(self):
        self.data = self.aead.seal(lzma.compress(self.data), self.nonce)

    def decrypt(self):
        self.data = lzma.decompress(self.aead.open(self.data, self.nonce))


class LZMA_AES_GCM_Block(Block, id='aes-gcm'):
    """ Another block implementation, using AES-GCM via the cryptography
    module. Mainly used to prove that independent block implementations
    work as intended.

    Blocks are compressed using LZMA before being encrypted.

    """

    __slots__ = ('password', 'nonce', 'data')

    def __init__(self, password, nonce, data):
        self.password = hashlib.sha256(password).digest()

        # use the NIST recommended 96 bits for the nonce
        self.nonce = hashlib.blake2b(nonce, digest_size=12).digest()

        self.data = data

    @property
    def aesgcm(self):
        return AESGCM(self.password)

    def encrypt(self):
        self.data = lzma.compress(self.data)
        self.data = self.aesgcm.encrypt(self.nonce, self.data, None)

    def decrypt(self):
        self.data = self.aesgcm.decrypt(self.nonce, self.data, None)
        self.data = lzma.decompress(self.data)



from logbook import Logger
from os import supports_follow_symlinks

log = Logger('spare')

FOLLOW_SYMLINKS = supports_follow_symlinks and {'follow_symlinks': True} or {}

__all__ = ('log', 'FOLLOW_SYMLINKS')
